from datetime import datetime
from meteostat import Hourly
import pulire as pu
from pulire.schema import Schema
from pulire.validator import ValidationError


df = Hourly("10637", datetime(2018, 1, 1), datetime(2018, 1, 1, 23, 59)).fetch()

myschema = Schema(
    {
        "temp": [pu.validators.minimum(-100), pu.validators.maximum(65)],
        "prcp": [pu.validators.minimum(0), pu.validators.maximum(350)],
        "snow": [pu.validators.minimum(0), pu.validators.maximum(11000)],
        "wdir": [pu.validators.minimum(0), pu.validators.maximum(360)],
        "wspd": [pu.validators.minimum(0), pu.validators.maximum(250)],
        "wpgt": [
            pu.validators.minimum(0),
            pu.validators.maximum(500),
            pu.validators.greater("wspd"),
        ],
        "tsun": [pu.validators.minimum(0), pu.validators.maximum(60)],
        "srad": [pu.validators.minimum(0), pu.validators.maximum(1368)],
        "pres": [pu.validators.minimum(850), pu.validators.maximum(1090)],
        "rhum": [pu.validators.minimum(0), pu.validators.maximum(100)],
        "cldc": [pu.validators.minimum(0), pu.validators.maximum(100)],
        "vsby": [pu.validators.minimum(0), pu.validators.maximum(9999)],
        "coco": [pu.validators.minimum(1), pu.validators.maximum(27)],
    }
)


try:
    myschema.debug(df)
except ValidationError as error:
    print(error.errors)


validated_df = myschema.validate(df)

print(validated_df)



"""
PULIRE

Validator Class

The code is licensed under the MIT license.
"""

from inspect import signature
from typing import Callable, Union
from pandas import DataFrame, Series


class Validator:
    """
    Series Validator
    """

    func: Union[Callable, None] = None
    vectorized: bool = True
    skip_null: bool = False

    def __init__(self, func: Callable, vectorized: bool = True, skip_null=True):
        self.func = func
        self.vectorized = vectorized
        self.skip_null = skip_null

    def check(self, df: DataFrame, column: str) -> Union[bool, Series]:
        """
        Run validator

        Returns a bool series:
        True -> Check passed
        False -> Check failed
        """
        arg_count = len((signature(self.func)).parameters)
        args = [df[column] if column else df, df, column]
        if self.vectorized:
            return self.func(*args[0:arg_count])
        raise NotImplementedError("Pulire doesn't support non-vectorized checks, yet.")


class ValidationError(Exception):
    """
    Exception which is raised when a validation fails
    """

    column: str = None
    validator: int = None
    errors: DataFrame = None

    def __init__(self, column: str, validator: int, df: DataFrame, test: Series):
        super().__init__(
            f"Validator [{validator}] failing for column '{column}' on the following rows:\n"
            f"{df[~test]}"
        )

        self.column = column
        self.validator = validator
        self.errors = df[~test]



"""
PULIRE

The code is licensed under the MIT license.
"""

from pandas import Series
from ..validator import Validator


def regex(pattern: str) -> Series:
    """
    Check if column values match regex pattern
    """
    return Validator(lambda s: s.str.contains(pattern))



"""
PULIRE

Schema Class

The code is licensed under the MIT license.
"""

from copy import copy
from inspect import isfunction
from pandas import DataFrame, Series
from .validator import ValidationError, Validator


class Schema:
    """
    Pulire Schema
    """

    _columns: dict = {}  # The columns

    def __init__(self, columns: dict):
        self._columns = columns

    def columns(self) -> list:
        """
        Get the list of column names
        """
        return list(self._columns.keys())

    @staticmethod
    def _run_check(validator: Validator, df: DataFrame, column: str = None) -> Series:
        validator = validator() if isfunction(validator) else validator
        if validator.skip_null:
            result = Series(data=True, index=df.index, dtype=bool)
            result.update(validator.check(df.loc[df[column].notnull()], column))
            return result.astype(bool)
        return validator.check(df, column)

    def validate(self, df: DataFrame, fill=None) -> DataFrame:
        """
        Validate a DataFrame
        """
        temp = copy(df)

        for name, validators in self._columns.items():
            if name in df.columns:
                for validator in validators:
                    test = self._run_check(validator, df, name)
                    temp.loc[~test, name] = fill

        return temp

    def debug(self, df: DataFrame) -> None:
        """
        Raise error when checks are failing
        """
        for name, validators in self._columns.items():
            if name in df.columns:
                for i, validator in enumerate(validators):
                    test = self._run_check(validator, df, name)
                    if not test.all():
                        raise ValidationError(name, i, df, test)

    def is_valid(self, df: DataFrame) -> bool:
        """
        Check if a DataFrame is valid
        """
        for name, validators in self._columns.items():
            if name in df.columns:
                for validator in validators:
                    test = self._run_check(validator, df, name)
                    if not test.all():
                        return False
        return True



"""
PULIRE

The code is licensed under the MIT license.
"""

from pandas import Series
from ..validator import Validator


def isin(values: list) -> Series:
    """
    Require column value to be in a list of values
    """
    return Validator(lambda s: s.isin(values))



"""
PULIRE

The code is licensed under the MIT license.
"""

from typing import Union
from pandas import Series
from ..validator import Validator


def max_diff(value: Union[int, float]) -> Series:
    """
    Maximum difference compared to previous value
    """

    def _func(s: Series):
        result = Series(data=0, index=s.index)
        result.update(s.iloc[1:].diff().notnull().abs())
        return result <= value

    return Validator(_func, skip_null=False)



"""
PULIRE

The code is licensed under the MIT license.
"""

from pandas import Series
from ..validator import Validator


def requires(column: str) -> Series:
    """
    Require another column not to be null
    """
    return Validator(lambda _, s: s[column].notna(), skip_null=False)



"""
PULIRE

The code is licensed under the MIT license.
"""

from typing import Union
from pandas import Series
from scipy.signal import find_peaks, peak_prominences
from ..validator import Validator


def max_peak(value: Union[int, float]) -> Series:
    """
    Maximum peak in a time series
    """

    def _func(series):
        result = Series(data=0, index=series.index)
        peaks, _ = find_peaks(series.values)
        prominences = peak_prominences(series.values, peaks, 2)[0]
        result.iloc[peaks] = prominences
        return result < value

    return Validator(_func, skip_null=False)



"""
PULIRE

The code is licensed under the MIT license.
"""

from typing import Union
from pandas import Series
from ..validator import Validator


def maximum(value: Union[int, float]) -> Series:
    """
    Numeric maximum
    """
    return Validator(lambda s: s <= value)
